#!/usr/bin/perl
#
#  Wedge Martin  
#  IBM     2007
#
#  Process binary AMA files and XML IPDR files to create CSV for
#  the UBA to process.
#
#  $Id: cs2kcdr.pl 5156 2008-06-20 05:13:12Z wmartin $
#  Service Changes - 4823/0.1 - Call codes changed to meet BT Call codes, BT Directories - J Slater 1/10/07
#       4823/0.2 - Amended from FTP to read directories - J Slater 28/10/07
#	4823/0.3 - Serv PMR fixes applied J Slater 081107 - config file created, sort and fine tune match of AMA/IPDR
#    
#  11/07 - wedge - Changed the way in which we set timestamps for aggregation intervals.  This was a 
#                  huge performance increase.
#  11/07 - wedge - Fixed a nasty issue with AMA file sorting.  We'd get the newest AMA first instead 
#                  of the oldest.  This caused a significant data loss issue.
#
#
# PDR format looks like     QCA.{nni}.QoS.{yyyy}.{mm}.{dd}_{hh}.{mm}_GMT.xml.gz QCA.9176524.QoS.2007.08.23_15.36_GMT.xml.gz
# AMA format looks like   {nni}_U{yy}{mm}{dd}{hh}{mm}{ss}AMA.unp  9176516_U070823024935AMA.unp
# keys
# nni  = the sensor ID for the switch - This is the same number that is contained in the AMA record.
# yyyy = 2007
# yy = 07
# mm = 08
# dd = 23
# hh = 13
# mm = 02
# ss = seconds
# unp = unprocessed

use strict;

use POSIX qw(setsid);


#use Net::FTP;
use Date::Manip;
use Data::Dumper;
use File::Copy;
use File::Basename;
use constant TRACK_EXTENSIONS => 1;


### AMA variables


my $totalRecords = 0;
my $logType;
my $lastDate;
my $recordHash;
my $currentRecord;
my $startEpoch;
my $launchpath = $0;
my $lastAMAstamp;


$SIG{'HUP'} = sub {
   open(TMP_HUP, "> /tmp/cs2k.dump");
   print TMP_HUP Dumper($recordHash);
   close(TMP_HUP);
};


##################################
###### USER CONFIG SECTION #######
##################################

# This has been moved to cs2kcdr-term.pl

BEGIN {
   my $pathname = dirname($0);
   require "$pathname/cs2kcdr-config.pl";
};

use vars qw( 
   $logDir $NNI $maxFilesPerRun $amaFilesPerRun 
   $runInterval $lagBehindDelay $globalAggrInterval
   $amaJunkRecords $numberOfOutputs $outputDir $exceptDir
   $storeExceptionDetail
   $terminationCodeHash $cdrConfig $trunkConfig $locationMap
   $amapath $amadir
   $ipdrpath $ipdrdir

);

##### END EDITABLE SECTION ####################
##### END EDITABLE SECTION ####################
##### END EDITABLE SECTION ####################


my $logTimestamp = `date +%Y%m%d-%H%M%S`;
chomp($logTimestamp);
my $logFile = "cs2kcdr_$logTimestamp.log";

close(STDIN);
close(STDOUT);
close(STDERR);
open(STDIN, '/dev/null');
open(STDOUT, "> $logDir/$logFile");
open(STDERR, "> $logDir/$logFile");




########################################
##### function _getSubelementName ######
########################################

sub _getSubelementName( $ ) {
     my $type = shift;
     my $se_name = "CDR_";
     if ( $type eq "Zone" ) {
          $se_name .= $currentRecord->{cs2kID} . "_$NNI" . "_Zone_" . _getZone();
     } elsif ( $type eq "Path" ) {
          $se_name .= $currentRecord->{cs2kID} . "_$NNI" .  "_Path_" . _getZone() . "_" . $currentRecord->{trunk_facility};
     } elsif ( $type eq "Device" ) {
          $se_name .= $currentRecord->{cs2kID} . "_$NNI" . "_Device";
     } elsif ( $type eq "TC" ) {
          $se_name .= $currentRecord->{cs2kID} . "_$NNI" . "_TC_";
          $se_name .= $terminationCodeHash->{$currentRecord->{facility_release_cause}} || $currentRecord->{facility_release_cause};
     } elsif ( $type eq "TypeOfCall" ) {
          $se_name .= $currentRecord->{cs2kID} . "_$NNI" . "_TypeOfCall_" . $currentRecord->{typeofcall};
     } elsif ( $type eq "Extension1" ) {
          $se_name .= "Extension" . "_$NNI" . "_" . _mapExtension($currentRecord->{originating_open_digit1}, $currentRecord->{originating_open_digit2});
     } elsif ( $type eq "Extension2" ) {
          $se_name .= "Extension" . "_$NNI" . "_" . _mapExtension($currentRecord->{terminating_open_digit1}, $currentRecord->{terminating_open_digit2});
     } elsif ( $type eq "Trunk" ) {
          $se_name .= $currentRecord->{cs2kID} . "_$NNI" . "_Trunk_" . _getTrunkId();
     } elsif ( $type eq "Customer" ) {
          my $profile;
          $profile = "External" if ( $currentRecord->{originating_charge_info} eq "001" );
          $profile = "Residential" if ( $currentRecord->{originating_charge_info} eq "010" );
          $profile = "Corporate" if ( $currentRecord->{originating_charge_info} eq "020" );
          $profile = "InterTrunk" if ( $currentRecord->{originating_charge_info} eq "255" );
          $se_name .= $currentRecord->{cs2kID} . "_$NNI" . "_Customer_" . $profile;
     }
     return $se_name;
}



#####################################
##### function _getElementName ######
#####################################

sub _getElementName() {
    my $e_name;
    if ( $cdrConfig->{$currentRecord->{sensor_id}} ) {
        $e_name = $cdrConfig->{$currentRecord->{sensor_id}}->{name};
    } elsif ( $currentRecord->{sensor_id} ) {
        $e_name = $currentRecord->{sensor_id};
    } else {
        $e_name = "SoftswitchUnknown";
    }
    return $e_name;
}



#####################################
##### function _getCS2Kid ######
#####################################

sub _getCS2Kid( $ ) {
    my $id = shift;
    if ( $cdrConfig->{$id} ) {
         return $cdrConfig->{$id}->{name};
    } else {
         return $id;
    }
}


#####################################
##### function _getTrunkId ######
#####################################

sub _getTrunkId() {
    my $trunkId = substr($currentRecord->{trunk_facility}, 1, 4);
    if ( $trunkConfig->{$trunkId} ) {
          return $trunkConfig->{$trunkId}->{name};
    } else {
          return $trunkId;
    }
}



sub _amastamp( $ ) {
  my $fname = shift;
  my $file;
  ( $file ) = $fname =~ /_U(.*)A/;
  my $year = substr( $file, 0, 2 );
  $year += 2000;
  my $month = substr( $file, 2, 2 );
  my $day = substr( $file, 4, 2 );
  my $hour = substr( $file, 6, 2 );
  my $minute = substr( $file, 8, 2 );
  my $second = substr( $file, 10, 2 );
  my $stamp;
  if ( $month ) {
     $stamp = Date_SecsSince1970( $month, $day, $year, $hour, $minute, $second );
  }
  return $stamp;
}


sub _amasort() {
   # AMA format looks like                 {nni}_U{yy}{mm}{dd}{hh}{mm}{ss}AMA.unp  9176516_U070823024935AMA.unp
    my $sorthash;
    my ( $a_file, $b_file );
    if ( $a && $b ) {
       $sorthash->{$a}->{timestamp} = _amastamp($a);
       $sorthash->{$b}->{timestamp} = _amastamp($b);
    }
    $sorthash->{$a}->{timestamp} <=> $sorthash->{$b}->{timestamp};
}


sub _ipdrstamp( $ ) {
   my $fname = shift;
   my $line = [ split(/\./, $fname ) ];
   my $year = $line->[3];
   my $month = $line->[4];
   my ( $day, $hour ) = split(/_/, $line->[5]);
   my ( $minute, undef ) = split(/_/, $line->[6]);
   my $stamp;
   if ( $month ) {
      $stamp = Date_SecsSince1970( $month, $day, $year, $hour, $minute, 0 );
   }
   return $stamp;
}

sub _ipdrsort() {
   my $sorthash;
   if ( $a && $b ) {
      # IPDR format looks like     QCA.{nni}.QoS.{yyyy}.{mm}.{dd}_{hh}.{mm}_GMT.xml.gz QCA.9176524.QoS.2007.08.23_15.36_GMT.xml.gz
      $sorthash->{$a}->{timestamp} = _ipdrstamp($a);
      $sorthash->{$b}->{timestamp} = _ipdrstamp($b);
   }
   
   $sorthash->{$a}->{timestamp} <=> $sorthash->{$b}->{timestamp};
  

}


#####################################
##### function _processAMA ######
#####################################

sub _processAMA() {

   $logType = "ama";
   my $newFileArr;

   opendir(AMA_STORE, $amadir) || die "Could not open dir '$amadir': $!\n";
   my @old_files = readdir(AMA_STORE);
   close(AMA_STORE);

   opendir(NEW_FILES, $amapath) || die "Could not open dir '$amapath': $!\n";
   my @predir = readdir(NEW_FILES);

   my $fileArray;
   my $fcounter = 0;
   # map { $_ = (split(/\s+/, $_)) } @predir;
   foreach my $listFile ( sort _amasort grep {$_ ne '.' && $_ ne '..' } @predir ) {
      push(@$fileArray, $listFile) if $listFile;
   }

   foreach my $f ( @$fileArray ) {
       if ( ! grep { $_ eq $f } @old_files ) {
            # we have a new file...
            if ( $fcounter < $amaFilesPerRun ) {
               # $ftp->binary();
               $lastAMAstamp = _amastamp($f);
               print "Downloading file: $f ... ";
               # move("$amapath/$f", "$amadir/$f");
               move("$amapath/$f", "$amadir/$f");
               print "Done.\n";
               push ( @$newFileArr, $f );
               $fcounter++;
            }
       } else {
            print "Skipping already obtained file: $f\n";
       }

   }


   my $timestampStart = time();
   if ( $newFileArr ) {
      print "Processing AMA files.... TIMESTAMP: $timestampStart\n";
   }
   foreach my $file ( @$newFileArr ) {
      my ( $data, $foo );
      open (AMAFILE, "< $amadir/$file") || die "Unable to open file '$file': $!\n";
      binmode(AMAFILE);
      while ( read(AMAFILE, $foo, 180000) ) {
          my $n1 = length($foo);
          my $s1 = 2 * $n1;
          my $datahere = unpack("H$s1", $foo);
          $datahere =~ tr/[a-z]/[A-Z]/;
          $data .= $datahere;
      }

      my $recordArr = [ split(/(?:AA|AB)+/, $data) ]; 
      my ( $callId, $timestamp, $dataArr, $date, $time, $year, $month, $day, $hour, $minute, $second, @dateArray, @timeArray );
      foreach my $record ( @$recordArr ) {
         $dataArr = [ split(/(?:FFFFFFFFFFFFFFFF|FFFFFFFFFF|FFFF|C)/, $record) ];
         next if ( scalar(@$dataArr) < 5 );
         next if ( ! grep { $_ eq "612" } @$dataArr );
         $date = $dataArr->[6];
         $time = $dataArr->[21];
         @dateArray = split(//, $date);
         $year = "200" . $dateArray[0];
         $month = $dateArray[1] . $dateArray[2];
         $day = $dateArray[3] . $dateArray[4];

         @timeArray = split(//, $time);
         $hour = $timeArray[0] . $timeArray[1];
         $minute = $timeArray[2] . $timeArray[3];
         # $second = $timeArray[4] . $timeArray[5];
         $second = "0";   # we don't do sub-minute granularity so we can take a load off of the script 
                           # by adjusting the timestamp to round the minute down.
         eval {
             $timestamp = Date_SecsSince1970($month,$day,$year,$hour,$minute,$second);
         };
         print "ERROR: $@\n" if ( $@ );
         my $trunk_facility;
         my $facility_release_cause;
         foreach ( my $marker = 22; $marker < 40; $marker++ ) {
             my $field = $dataArr->[$marker];
             if ( $field eq "612" && $dataArr->[$marker + 1] eq "8006600" ) {
                   $callId = $dataArr->[$marker + 2] . $dataArr->[$marker + 3];
             } elsif ( $field eq "104" ) {
                   $trunk_facility = $dataArr->[$marker + 1];
             } elsif ( $field eq "130" ) {
                   $facility_release_cause = substr($dataArr->[$marker + 1], 2, 5);
             }
         }
         $callId = _generateCallId( $callId );
         # print "CALLID: $callId\n";


         $totalRecords++;
         $recordHash->{$callId}->{ama} = 1;
         $recordHash->{$callId}->{timestamp} = $timestamp;
         $recordHash->{$callId}->{structure_code} = $dataArr->[0];
         $recordHash->{$callId}->{call_code} = $dataArr->[1];
         $recordHash->{$callId}->{sensor_type} = $dataArr->[2];
         $recordHash->{$callId}->{sensor_id} = $dataArr->[3];
         $recordHash->{$callId}->{office_type} = $dataArr->[4];
         $recordHash->{$callId}->{office_id} = $dataArr->[5];
         $recordHash->{$callId}->{UNUSED_date} = $dataArr->[6];
         $recordHash->{$callId}->{timing_indicator} = $dataArr->[7];
         $recordHash->{$callId}->{study_indicator} = $dataArr->[8];
         $recordHash->{$callId}->{answer} = $dataArr->[9];
         $recordHash->{$callId}->{service_observed} = $dataArr->[10];
         $recordHash->{$callId}->{operator_action} = $dataArr->[11];
         $recordHash->{$callId}->{service_feature} = $dataArr->[12];
         $recordHash->{$callId}->{sig_digits_caller} = $dataArr->[13];
         $recordHash->{$callId}->{originating_open_digit1} = $dataArr->[14];
         $recordHash->{$callId}->{originating_open_digit2} = $dataArr->[15];
         $recordHash->{$callId}->{originating_charge_info} = $dataArr->[16];
         $recordHash->{$callId}->{domestic_intl_indicator} = $dataArr->[17];
         $recordHash->{$callId}->{sig_digits_callee} = $dataArr->[18];
         $recordHash->{$callId}->{terminating_open_digit1} = $dataArr->[19];
         $recordHash->{$callId}->{terminating_open_digit2} = $dataArr->[20];
         $recordHash->{$callId}->{connect_time} = $dataArr->[21];
         $recordHash->{$callId}->{elapsed_time} = $dataArr->[22];
         if ( !  $recordHash->{$callId}->{answer}  || $recordHash->{$callId}->{answer} == 0 ) {
               $recordHash->{$callId}->{answer} = 1;
         } else {
               $recordHash->{$callId}->{answer} = 2;
         }
         $recordHash->{$callId}->{cs2kID} = _getCS2Kid($recordHash->{$callId}->{sensor_id});
         $recordHash->{$callId}->{trunk_facility} = $trunk_facility;
         $recordHash->{$callId}->{facility_release_cause} = $facility_release_cause;
         my $trunk_type = substr($trunk_facility, 0, 1);
         my $ctype = $recordHash->{$callId}->{call_code};
         if ( $ctype eq "804" ) {
             $ctype = "Internal";
         } elsif ( $trunk_type == 1 ) { 
             $ctype = "Incoming";
         } elsif ( $trunk_type == 2 ) { 
             $ctype = "Outgoing";
         }
         $recordHash->{$callId}->{typeofcall} = $ctype;

       }
       close(AMAFILE);
   }

   
}



sub _currentCallIsComplete() {
    # print "DUMPER_in_currentCallIsComplete(): " . Dumper($currentRecord, undef) . "\n\n";
    if ( $currentRecord->{ama} == 1 && $currentRecord->{ipdr} == 1 ) {
        return 1;
    }
}


sub _getZone() {
   if ( $currentRecord->{sensor_id} ) {
       if ( $cdrConfig->{$currentRecord->{sensor_id}} ) {
          return $cdrConfig->{$currentRecord->{sensor_id}}->{zone};
       } else {
          return $currentRecord->{sensor_id};
      }
   } elsif ( $currentRecord->{subscriberId} ) {
      if ( $currentRecord->{subscriberId} =~ /aaln/ ) {
           if ( $currentRecord->{subscriberId} =~ /\./ ) {
              my $subscriber = (split(/\./, $currentRecord->{subscriberId}))[0];
              return $subscriber;
          } else {
              return $currentRecord->{subscriberId};
          }
      }
   }
}


sub _getSoftswitchLocation( $ ) {
   if ( $cdrConfig->{$currentRecord->{sensor_id}} ) {
        return $cdrConfig->{$currentRecord->{sensor_id}}->{location};
   }
   return "Unmapped";
}


sub _getLocation( $ $ ) {
   my ( $data, $type ) = @_;
   my $stripDigits;
   return $data if ( $data eq "Unknown" );
   foreach my $loc ( keys %$locationMap ) {
       if ( $type eq "extension" ) { 
           my $regexp = $locationMap->{$loc}->{extensionPattern};
           if ( $regexp && $data =~ $locationMap->{$loc}->{extensionPattern} ) {
                ( $stripDigits ) = $data =~ /$regexp(.*)/;
                if ( $locationMap->{$loc}->{extensionRange} ) {
                     return $loc;
                }
                my @stripDigitArr = split(//, $stripDigits);
                for ( my $r = 0; $r < scalar(@{$locationMap->{$loc}->{extensionRange}}); $r++ ) {
                     my $range = $locationMap->{$loc}->{extensionRange}->[$r];
                     my $sizedLastDigits;
                     if ( $range =~ /-/ ) {
                         my @nums = split(//, $range);
                         for ( my $i = 0; $i < length($nums[0]); $i++ ) {
                             $sizedLastDigits .= $stripDigitArr[$i];
                         }
                         if ( $sizedLastDigits >= $nums[0] && $sizedLastDigits <= $nums[1] ) {
                               return $loc;
                         }
                     } else {
                         for ( my $i = 0; $i < length($range); $i++ ) {
                               $sizedLastDigits .= $stripDigitArr[$i];
                         }
                         if ( $sizedLastDigits eq $range ) {
                               return $loc;
                         }
                     }
                }
           }
      } elsif ( $type eq "ipAddress" ) {
           my $returnLoc = "false";
           my $ip = $data;
           if ( $locationMap->{$loc}->{ipAddress} ) {
               my @maskArray = split(/\./, $locationMap->{$loc}->{ipAddress});
               my @ipArray = split(/\./, $ip);
               for ( my $p = 0; $p < scalar(@maskArray); $p++ ) {
                   if ( $maskArray[$p] =~ /-/ ) {
                       my @stoppers = split(/-/, $maskArray[$p]);
                       if ( $ipArray[$p] >= $stoppers[0] && $ipArray[$p] <= $stoppers[1] ) {
                            $returnLoc = "true";
                       }
                   } else {
                       if ( $ipArray[$p] == $maskArray[$p] ) {
                            $returnLoc = "true";
                       } else {
                            $returnLoc = "false";
                       }
                   }
               }  
           }
           if ( $returnLoc eq "true" ) {
               return $loc;
           }
      }
   }
   return "Unmapped";
}




sub _mapExtension( $ $ ) {
   my ( $field1, $field2 ) = @_;
   my $extension;
   if ( ! $field1 && ! $field2 ) {
        return "Untracked";
   }
   my $check;
   $extension = $field1;
   if ( $field1 !~ /^0/ && $field2 ) {
      $extension .= $field2;
   }
   my $locTest = _getLocation($extension, "extension");
   if ( $locTest !~ /Unknown/ && $locTest !~ /Unmapped/ ) {
       return $locTest;
   } else {
       return "Untracked";
   }
}



sub _isOffNetCall() {
    if ( $currentRecord->{trunk_facility} ) {
         return 1;
    } else {
         return 0;
    }
}

sub _isCallCode( $ ) {
   my $cc = shift;
   if ( $currentRecord->{call_code} eq $cc ) {
       return 1;
   } else {
       return 0;
   }
}

sub _getElapsedTime() {
     if ( $currentRecord->{elapsed_time} ) {
         return $currentRecord->{elapsed_time};
     }
}


sub _isShortCall() {
    if ( 2 > $currentRecord->{elapsed_time} ) {
        return 1;
    }
}


sub _isSuccessfulCall() {
    if ( $currentRecord->{answer} == 1 ) {
        return 1;
    }
}


sub _isUnsuccessfulCall() {
   if ( $currentRecord->{answer} == 2 ) {
        return 1;
   }
}


sub _isFacilityReleaseCause( $ ) {
    my $frc = shift;
    if ( $currentRecord->{facility_release_cause} eq $frc ) {
         return 1;
    }
}

sub _isCallCompletion( $ ) {
    my $cc = shift;
    if ( $currentRecord->{call_completion} == $cc ) {
        return 1;
    }
}


sub _getPacketLatency() {
   return $currentRecord->{averagePacketLatency};
}

sub _getInboundPackets() {
   return $currentRecord->{inboundPacketCount};
}

sub _getOutboundPackets() {
   return $currentRecord->{outboundPacketCount};
}

sub _getInboundBytes() {
   return $currentRecord->{inboundByteCount};
}

sub _getOutboundBytes() {
   return $currentRecord->{outboundByteCount};
}

sub _getLostPackets() {
   return $currentRecord->{inboundLostPacketCount};
}

sub _getPacketDelayVariation() {
   return $currentRecord->{packetDelayVariation};
}




my $functionMap = {

    "TotalCall" =>                      { aggr => "sum", type =>  "ama",  method =>  sub { return 1 if $currentRecord->{ama} == 1 } },
    # "IncomingCall890" =>                   { aggr => "sum", type =>  "ipdr", method =>  sub { return _isCallCode(890) } },
    # "IncomingCall891" =>                   { aggr => "sum", type =>  "ipdr", method =>  sub { return _isCallCode(891) } },
    "IncomingCall" =>                   { aggr => "sum", type =>  "ipdr", method =>  sub { return ( _isCallCode(891) || _isCallCode(890) ) } },
    "OutgoingCall" =>                   { aggr => "sum", type =>  "ipdr", method =>  sub { return _isCallCode(6) } },
    "OnNetCall" =>                      { aggr => "sum", type =>  "ipdr", method =>  sub { return _isCallCode(901) } },
    "OffNetCall" =>                     { aggr => "sum", type =>  "ama",  method =>  sub { return _isOffNetCall() } },
    "InternalCall" =>                   { aggr => "sum", type =>  "ipdr", method =>  sub { return _isCallCode(901) } },
    "MobileCall" =>                     { aggr => "sum", type =>  "ipdr", method =>  sub { return _isCallCode(801) } },
    "InternationalCall" =>              { aggr => "sum", type =>  "ipdr", method =>  sub { return _isCallCode(800) } },
    "NationalCall" =>                   { aggr => "sum", type =>  "ipdr", method =>  sub { return _isCallCode(805) } },
    "SuccessfulCall" =>                 { aggr => "sum", type =>  "ama",  method =>  sub { return _isSuccessfulCall() } },
    "UnsuccessfulCall" =>               { aggr => "sum", type =>  "ama",  method =>  sub { return _isUnsuccessfulCall() } },
    "UnsuccessfulCallInvalidAddress" => { aggr => "sum", type =>  "ama",  method =>  sub { return _isFacilityReleaseCause("517") } },
    "UnsuccessfulCallNoAnswer" =>       { aggr => "sum", type =>  "ama",  method =>  sub { return _isFacilityReleaseCause("012") } },
    "UnsuccessfulCallBusy" =>           { aggr => "sum", type =>  "ama",  method =>  sub { return _isFacilityReleaseCause("001") } },
    "UnsuccessfulCallNetwork" =>        { aggr => "sum", type =>  "ama",  method =>  sub { return _isFacilityReleaseCause("002") } },
    "CallDuration" =>                   { aggr => "average", type =>  "ama",  method =>  sub { return _getElapsedTime() } },
    "NormalCallTermination" =>          { aggr => "sum", type =>  "ipdr", method =>  sub { return _isCallCompletion("CC") } },
    "DropCall" =>                       { aggr => "sum", type =>  "ipdr", method =>  sub { return _isCallCompletion("CAD") } },
    "ShortCall" =>                      { aggr => "sum", type =>  "ama",  method =>  sub { return _isShortCall() } },
    "PacketLatency" =>                  { aggr => "average", type =>  "ipdr", method =>  sub { return _getPacketLatency() } },
    "InboundPacket" =>                  { aggr => "sum", type =>  "ipdr", method =>  sub { return _getInboundPackets() } },
    "OutboundPacket" =>                 { aggr => "sum", type =>  "ipdr", method =>  sub { return _getOutboundPackets() } },
    "InboundByte" =>                    { aggr => "sum", type =>  "ipdr", method =>  sub { return _getInboundBytes() } },
    "OutboundByte" =>                   { aggr => "sum", type =>  "ipdr", method =>  sub { return _getOutboundBytes() } },
    "LostPacket" =>                     { aggr => "sum", type =>  "ipdr", method =>  sub { return _getLostPackets() } },
    "PacketDelayVariation" =>           { aggr => "average", type =>  "ipdr", method =>  sub { return _getPacketDelayVariation() } }
};



my $deviceMetrics = {
      Trunk =>       { formulaPath =>  "AP~Specific~Bulk~Carrier VOIP~Infrastructure~Class 5 CDR~CDR Media Gateway", 
                    filter =>  sub { return 1 if ( length($currentRecord->{trunk_facility}) > 4 || _currentCallIsComplete() ) },
                    properties =>  { VOIPSoftswitchComponentType =>  "CDR_MGW", VOIPSoftswitchVendor =>  "Nortel", 
                                  VOIPSoftswitchZone =>  sub { return _getZone() }, 
                                  VOIPSoftswitchLocation =>  sub { return _getSoftswitchLocation($currentRecord->{sensor_id}) } },
                    metricArray =>  [ "TotalCall", "IncomingCall", "OutgoingCall", "OnNetCall", "OffNetCall", "InternalCall", "MobileCall", 
                                   "InternationalCall", "NationalCall", "SuccessfulCall", "UnsuccessfulCall", "UnsuccessfulCallInvalidAddress", 
                                   "UnsuccessfulCallNoAnswer", "UnsuccessfulCallBusy", "UnsuccessfulCallNetwork",
                                   "DropCall", "CallDuration", "ShortCall" ]
                  },
      TypeOfCall =>  { formulaPath =>  "AP~Specific~Bulk~Carrier VOIP~Infrastructure~Class 5 CDR~CDR Type of Call", 
                    filter =>  sub { return 1 if (  ( $currentRecord->{call_code} > 0 && $currentRecord->{sensor_id} > 0 ) || _currentCallIsComplete() ) },
                    properties =>  { VOIPSoftswitchComponentType =>  "CDR_TypeOfCall", VOIPSoftswitchVendor =>  "Nortel", 
                                  VOIPSoftswitchZone =>  sub { return _getZone() }, 
                                  VOIPSoftswitchLocation =>  sub { return _getSoftswitchLocation($currentRecord->{sensor_id}) } },
                    metricArray =>  [ "TotalCall", "MobileCall", "SuccessfulCall", "UnsuccessfulCall", "UnsuccessfulCallInvalidAddress",
                                   "UnsuccessfulCallNoAnswer", "UnsuccessfulCallBusy", "UnsuccessfulCallNetwork",
                                   "DropCall", "CallDuration", "ShortCall" ]
                  },
      Customer =>    { formulaPath =>  "AP~Specific~Bulk~Carrier VOIP~Infrastructure~Class 5 CDR~CDR Customer Profile",
                    filter =>  sub { return 1 if ( ( $currentRecord->{call_code} > 0 && $currentRecord->{sensor_id} > 0 && $currentRecord->{originating_charge_info} > 0 ) 
                                       || _currentCallIsComplete() ) },
                    properties =>  { VOIPSoftswitchComponentType =>  "CDR_Softswitch_CustomerProfile", VOIPSoftswitchVendor =>  "Nortel", 
                                  VOIPSoftswitchZone =>  sub { return _getZone() }, 
                                  VOIPSoftswitchLocation =>  sub { return _getSoftswitchLocation($currentRecord->{sensor_id}) } },
                    metricArray =>  [ "TotalCall", "IncomingCall", "OutgoingCall", "OnNetCall", "OffNetCall", "InternalCall", "MobileCall", 
                                   "InternationalCall", "NationalCall", "SuccessfulCall", "UnsuccessfulCall", "UnsuccessfulCallInvalidAddress", 
                                   "UnsuccessfulCallNoAnswer", "UnsuccessfulCallBusy", "UnsuccessfulCallNetwork", "NormalCallTermination", 
                                   "DropCall", "CallDuration", "PacketLatency", "InboundPacket", "OutboundPacket", "InboundByte", "OutboundByte", 
                                   "LostPacket", "PacketDelayVariation", "ShortCall" ]
                  },
      Path =>        { formulaPath =>  "AP~Specific~Bulk~Carrier VOIP~Infrastructure~Class 5 CDR~CDR Path",
                    filter =>  sub { return 1 if ( _currentCallIsComplete() ) },
                    properties =>  { VOIPSoftswitchComponentType =>  "CDR_Path", VOIPSoftswitchVendor =>  "Nortel", 
                                  VOIPSoftswitchZone =>  sub { return _getZone() }, 
                                  VOIPSoftswitchLocation =>  sub { return _getSoftswitchLocation($currentRecord->{sensor_id}) } },
                    metricArray =>  [ "TotalCall", "MobileCall", "InternationalCall", "NationalCall", "SuccessfulCall", "UnsuccessfulCall",
                                   "UnsuccessfulCallInvalidAddress", "UnsuccessfulCallNoAnswer", "UnsuccessfulCallBusy", "UnsuccessfulCallNetwork",
                                   "NormalCallTermination", "DropCall", "CallDuration", "PacketLatency", "InboundPacket", "OutboundPacket",
                                   "InboundByte", "OutboundByte", "LostPacket", "PacketDelayVariation", "ShortCall" ]
                  },
      Zone =>        { formulaPath =>  "AP~Specific~Bulk~Carrier VOIP~Infrastructure~Class 5 CDR~CDR Zone",
                    filter =>  sub { return 1 if ( _currentCallIsComplete() ) },
                    properties =>  { VOIPSoftswitchComponentType =>  "CDR_Zone", VOIPSoftswitchVendor =>  "Nortel", 
                                  VOIPSoftswitchZone =>  sub { return _getZone() }, 
                                  VOIPSoftswitchLocation =>  sub { return _getSoftswitchLocation($currentRecord->{sensor_id}) } },
                    metricArray =>  [ "TotalCall", "MobileCall", "InternationalCall", "NationalCall", "SuccessfulCall", "UnsuccessfulCall",
                                   "UnsuccessfulCallInvalidAddress", "UnsuccessfulCallNoAnswer", "UnsuccessfulCallBusy", "UnsuccessfulCallNetwork",
                                   "NormalCallTermination", "DropCall", "CallDuration", "PacketLatency", "InboundPacket", "OutboundPacket",
                                   "InboundByte", "OutboundByte", "LostPacket", "PacketDelayVariation", "ShortCall" ]
                  },
      Device =>      { formulaPath =>  "AP~Specific~Bulk~Carrier VOIP~Infrastructure~Class 5 CDR~CDR Softswitch Device",
                    filter =>  sub { return 1 },
                    properties =>  { VOIPSoftswitchComponentType =>  "CDR_Softswitch_Device", VOIPSoftswitchVendor =>  "Nortel", 
                                  VOIPSoftswitchZone =>  sub { return _getZone() }, 
                                  VOIPSoftswitchLocation =>  sub { return _getSoftswitchLocation($currentRecord->{sensor_id}) } },
                    metricArray =>  [ "TotalCall", "IncomingCall", "OutgoingCall", "OnNetCall", "OffNetCall", "InternalCall", "MobileCall",
                                   "InternationalCall", "NationalCall", "SuccessfulCall", "UnsuccessfulCall", "UnsuccessfulCallInvalidAddress",
                                   "UnsuccessfulCallNoAnswer", "UnsuccessfulCallBusy", "UnsuccessfulCallNetwork", "NormalCallTermination",
                                   "DropCall", "CallDuration", "PacketLatency", "InboundPacket", "OutboundPacket", "InboundByte", "OutboundByte",
                                   "LostPacket", "PacketDelayVariation", "ShortCall" ]
                  },
      TC =>          { formulaPath =>  "AP~Specific~Bulk~Carrier VOIP~Infrastructure~Class 5 CDR~CDR Termination Cause",
                    filter =>  sub { return 1 if ( $currentRecord->{facility_release_cause} ) },
                    properties =>  { VOIPSoftswitchComponentType =>  "CDR_TC", VOIPSoftswitchVendor =>  "Nortel", 
                                  VOIPSoftswitchZone =>  sub { return _getZone() }, 
                                  VOIPSoftswitchLocation =>  sub { return _getSoftswitchLocation($currentRecord->{sensor_id}) } },
                    metricArray =>  [ "TotalCall", "IncomingCall", "OutgoingCall", "OnNetCall", "OffNetCall", "InternalCall", "MobileCall",
                                   "InternationalCall", "NationalCall", "ShortCall" ]
                  },
     Extension1 =>   { formulaPath =>  "AP~Specific~Bulk~Carrier VOIP~Infrastructure~Class 5 CDR~CDR Extension",
                filter =>  sub { return 1 if ( TRACK_EXTENSIONS && ( $logType == "ama" || ( $logType == "ipdr" &&  _currentCallIsComplete() ) ) ) },
                    properties =>  { VOIPSoftswitchComponentType =>  "CDR_Extension", VOIPSoftswitchVendor =>  "Nortel",
                                  VOIPSoftswitchZone =>  sub { return _getZone() },
                                  VOIPSoftswitchLocation =>  sub { return _getLocation(_mapExtension($currentRecord->{originating_open_digit1}, $currentRecord->{originating_open_digit2} ), "extension" ) } },
                    metricArray =>  [ "TotalCall", "IncomingCall", "OutgoingCall", "OnNetCall", "OffNetCall", "InternalCall", "MobileCall",
                                   "InternationalCall", "NationalCall", "SuccessfulCall", "UnsuccessfulCall", "UnsuccessfulCallInvalidAddress",
                                   "UnsuccessfulCallNoAnswer", "UnsuccessfulCallBusy", "UnsuccessfulCallNetwork", "NormalCallTermination",
                                   "DropCall", "CallDuration", "PacketLatency", "InboundPacket", "OutboundPacket", "InboundByte", "OutboundByte",
                                   "LostPacket", "PacketDelayVariation", "ShortCall" ]
                  },
     Extension2 =>   { formulaPath =>  "AP~Specific~Bulk~Carrier VOIP~Infrastructure~Class 5 CDR~CDR Extension",
                filter =>  sub { return 1 if ( TRACK_EXTENSIONS && ( $logType == "ama" || ( $logType == "ipdr" &&  _currentCallIsComplete() ) ) ) },
                    properties =>  { VOIPSoftswitchComponentType =>  "CDR_Extension", VOIPSoftswitchVendor =>  "Nortel",
                                  VOIPSoftswitchZone =>  sub { return _getZone() },
                                  VOIPSoftswitchLocation =>  sub { return _getLocation(_mapExtension($currentRecord->{terminating_open_digit1}, $currentRecord->{terminating_open_digit2} ), "extension" ) } },
                    metricArray =>  [ "TotalCall", "IncomingCall", "OutgoingCall", "OnNetCall", "OffNetCall", "InternalCall", "MobileCall",
                                   "InternationalCall", "NationalCall", "SuccessfulCall", "UnsuccessfulCall", "UnsuccessfulCallInvalidAddress",
                                   "UnsuccessfulCallNoAnswer", "UnsuccessfulCallBusy", "UnsuccessfulCallNetwork", "NormalCallTermination",
                                   "DropCall", "CallDuration", "PacketLatency", "InboundPacket", "OutboundPacket", "InboundByte", "OutboundByte",
                                   "LostPacket", "PacketDelayVariation", "ShortCall" ]
                  }

};



sub _checkExceptionCallDetail() {
    foreach my $code ( @{$storeExceptionDetail} ) {
        if ( $currentRecord->{facility_release_cause} == $code ) {
           return &UnixDate("epoch " . $currentRecord->{timestamp}, "%Y%m%d-%H%M%S");
        }
    }
}



sub _flushData() {

    my $datafile;
    my $lagfile;
    my $last_timestamp;
    my $exception_time;

    my $filecounter = 1;
    print "TotalRecords variable reports: $totalRecords\n";
    print "Beginning data flush to file... " . time() . "\n";

    # print "======\nCALLS: " . Dumper($recordHash, undef) . "\n\n";
    

    my $aggrHash;
    my $startAggrTime;
    my $aggrStop = "false";
    my $exceptionFile;

    if ( -f "/tmp/.debug" ) {
       unlink("/tmp/.debug");
    }
    open(DEBUG_OUT, "> /tmp/.debug");

    foreach my $r ( sort { $recordHash->{$a}->{timestamp} <=> $recordHash->{$b}->{timestamp} } keys %$recordHash ) {
          $currentRecord = $recordHash->{$r};

          my $time_complete;
          if ( $currentRecord->{timestamp} eq $lastDate->{stamp} ) {
              $time_complete = $lastDate->{time_complete};
          } elsif ( $currentRecord->{timestamp} ) {
              $time_complete = &UnixDate("epoch " . $currentRecord->{timestamp}, "%Y-%m-%d %H:%M:%S");
              if ( ! $currentRecord->{lagBehind} && $currentRecord->{timestamp} > $lastDate->{stamp} ) {
                 $lastDate->{stamp} = $currentRecord->{timestamp};
                 $lastDate->{time_complete} = $time_complete;
              }
          }

          print DEBUG_OUT $recordHash->{$r}->{timestamp} . " must be less than " . ( $lastDate->{stamp} - $lagBehindDelay ) . "\n";

          if ( $recordHash->{$r}->{timestamp} < ( $lastDate->{stamp} - $lagBehindDelay ) ) {
             #
             # record is old.  Go ahead and flush it out, but tag it to be handled by the stray handler.
             if ( $currentRecord->{ama} == 1 ) {
                 $currentRecord->{lagBehind} = 1;
             } else {
                 # print "Deleting key $r\n";
                 delete $recordHash->{$r};
             }
          }
          if ( ! _currentCallIsComplete() ) { 
             # print "Call is not complete...\n";
             next if ( ! $currentRecord->{lagBehind} );
          }
          if ( $storeExceptionDetail ) {
              my $exception_time = _checkExceptionCallDetail();
              if ( $exception_time ) {
                  if ( ! $exceptionFile ) {
		      #Amended exceptions output directory J Slater
                      open (EXCEPTIONS, "> $exceptDir/exception_" . "$NNI" . "_log_$exception_time");
                      $exceptionFile = \*EXCEPTIONS;
                  }
                  my $d = &UnixDate("epoch " . $currentRecord->{timestamp}, "%Y-%m-%d %H:%M:%S");
                  print $exceptionFile $d . "," . $r . "," . $currentRecord->{facility_release_cause} . "," 
                         . "," . $currentRecord->{call_code} . "," . $currentRecord->{elapsed_time} . "," 
                         . $currentRecord->{trunk_facility} . "," . $currentRecord->{sensor_id} . "," . $currentRecord->{answer} . "\n";
              }
          }

 
          $startAggrTime ||= $currentRecord->{timestamp};
          my $element = _getElementName();
          my $metric;
          my $aggr;
          my $subelement;
          my $val;


          ## new aggregation stuff..
          if ( $currentRecord->{timestamp} < ( $startAggrTime + $globalAggrInterval ) ) {
               $currentRecord->{timestamp} = $startAggrTime;
          } else {
               $startAggrTime = $currentRecord->{timestamp};
               $aggrStop = "true";
          }
           
          ## lastDate->{stamp} update was here...


          foreach my $deviceType ( keys %$deviceMetrics ) {
                if ( $deviceMetrics->{$deviceType}->{filter}->() ) {
                     $subelement = _getSubelementName($deviceType);
                     if ( $subelement !~ /Unknown/ && $subelement ) {
                          my $properties;
                          foreach my $prop ( keys %{$deviceMetrics->{$deviceType}->{properties}} ) {
                              my $check = sprintf($deviceMetrics->{$deviceType}->{properties}->{$prop});
                              if ( $check =~ /CODE/ ) {
                                  # property is defined as a subroutine.
                                  $properties .= $prop . "=" . $deviceMetrics->{$deviceType}->{properties}->{$prop}->() . "&";
                              } else {
                                  $properties .= $prop . "=" . $deviceMetrics->{$deviceType}->{properties}->{$prop} . "&";
                              }
                          }
                          for ( my $d = 0; $d < scalar(@{$deviceMetrics->{$deviceType}->{metricArray}}); $d++ ) {
                              $val = $functionMap->{$deviceMetrics->{$deviceType}->{metricArray}->[$d]}->{method}();
                              if ( $val ) {
                                  $metric = $deviceMetrics->{$deviceType}->{formulaPath} . "~" . $deviceMetrics->{$deviceType}->{metricArray}->[$d];
                                  $aggr = $functionMap->{$deviceMetrics->{$deviceType}->{metricArray}->[$d]}->{aggr};
                                  $last_timestamp = $currentRecord->{timestamp};
                                  push(@{$aggrHash->{$time_complete}->{$element}->{$subelement}->{$properties}->{$metric}->{values}}, $val);
                                  $aggrHash->{$time_complete}->{$element}->{$subelement}->{$properties}->{$metric}->{aggr} = $aggr;
                                  if ( $currentRecord->{lagBehind} ) {
                                       $aggrHash->{$time_complete}->{$element}->{$subelement}->{$properties}->{$metric}->{lagBehind} = 1;
                                  }
                              }
                          }
                     }
                }
          }
          # print "Removing key: " . $recordHash->{$r} . "\n";
          delete $recordHash->{$r};
    }

    close(DEBUG_OUT);

    print "Performing aggregations...\n";
  
    foreach my $time_c ( sort keys %$aggrHash ) {
        foreach my $ele ( keys %{$aggrHash->{$time_c}} ) {
             foreach my $se ( keys %{$aggrHash->{$time_c}->{$ele}} ) {
                  foreach my $prop ( keys %{$aggrHash->{$time_c}->{$ele}->{$se}} ) {
                       foreach my $metr ( keys %{$aggrHash->{$time_c}->{$ele}->{$se}->{$prop}} ) {
                            my ( $aggr, $val );

                            # print "Performing aggregation for: $time_c\n";
                            # print "        metric: $metr\n";
                            # print "        se: $se\n";
                            # print "        ele: $ele\n";

                            ##  do the aggregations
                            my $vals = $aggrHash->{$time_c}->{$ele}->{$se}->{$prop}->{$metr}->{values};
                            my $numEntries = scalar(@{$vals});
                            my $sumval;
                            $aggr = $aggrHash->{$time_c}->{$ele}->{$se}->{$prop}->{$metr}->{aggr};
                            if ( $aggr eq "average" ) {
                                 my $tmpval;
                                 map { $tmpval += $_ } @$vals;
                                 $val = $tmpval / $numEntries;
                            } else {
                                 map { $val += $_ } @$vals;
                            }
                            ## $aggrHash->{$time_complete}->{$element}->{$subelement}->{$properties}->{$metric}->{value} = $newval;
                            ##

                            if ( $aggrHash->{$time_c}->{$ele}->{$se}->{$prop}->{$metr}->{lagBehind} ) {
                               $lagfile .= join (',', ( $time_c,$ele,$se,$prop,$metr,$aggr,$val )) . "\n";
                            } else {
                               $datafile->[$filecounter - 1] .= join (',', ( $time_c,$ele,$se,$prop,$metr,$aggr,$val )) . "\n";
                               if ( $filecounter == $numberOfOutputs ) {
                                    $filecounter = 1;
                               } else {
                                    $filecounter++;
                               }
                            }
                       }
                  }
             }
        }
        delete $aggrHash->{$time_c};
    }

    print "Aggregations complete...\n";

    if ( $datafile ) {
       my $time = &UnixDate("epoch $last_timestamp", "%Y_%m_%d_%H_%M_%S");
       for ( my $f = 0; $f < scalar(@$datafile); $f++ ) {
          print "Flushing to file: $outputDir/CS2K_$time.$f.csv at " . time() . "\n";
          open(OUTPUTFILE, "> $outputDir/CS2K_$time.$f.csv");
          print OUTPUTFILE $datafile->[$f];
          close(OUTPUTFILE);
       }
    }
    if ( $lagfile ) {
       my $time = &UnixDate("epoch " . ( $last_timestamp - $lagBehindDelay ), "%Y_%m_%d_%H_%M_%S");
       print "Flushing to file: $outputDir/CS2K_$time.STRAY.csv at " . time() . "\n";
       open(LAGFILE, "> $outputDir/CS2K_$time.STRAY.csv");
       print LAGFILE $lagfile;
       close(LAGFILE);
    }
    my $ama_counter;
    foreach my $old_record ( keys %$recordHash ) {
       if ( $recordHash->{$old_record}->{ama} ) {
            $ama_counter++;
       }
    }
    print "TIMESTAMPEND: " . time() . "\n";
    print "Number of records still in memory after flush: " . scalar(keys %$recordHash) . "\n";
    print "Number of AMA records still in memory after flush: $ama_counter\n";
}







sub _generateCallId( $ ) {
     my $call = shift;
     my $hexCall; 
     my $callArr = [ split(//, $call) ];
     for ( my $i = 0; $i < length($call); $i = $i + 3 ) {
          my $hex = sprintf("%x", $callArr->[$i] . $callArr->[$i + 1] . $callArr->[$i + 2]);
          if ( length($hex) == 1 ) {
            $hex = "0" . $hex;
          }
          $hexCall .= $hex;
     }
     return $hexCall;
}




#########################################
##### IPDR handling


sub _processIPDR() {

   $logType = "ipdr";
   my $newFileArr;

   opendir(IPDR_STORE, $ipdrdir) || die "Could not open dir '$ipdrdir': $!\n";
   my @old_files = readdir(IPDR_STORE);

   #foreach my $host ( keys %$xmlHosts ) {
   #    print "Getting IPDR files from URI: $host ...\n";
   #   my $ftp = new Net::FTP( $host );
   #  $ftp->login( $xmlHosts->{$host}->{user}, $xmlHosts->{$host}->{pass} );
   # $ftp->cwd( $xmlHosts->{$host}->{path} );
   # my $dir = $ftp->dir();
   
   opendir(NEW_FILE, $ipdrpath) || die "Could not open dir '$ipdrpath': $!\n";
   my @dir = readdir(NEW_FILE);
   #my $dir = \@dirip ;
   #print "FOO: $dir. \n";
   my $fileArray;
   my $fcounter = 0;
   # map { $_ = (split(/\s+/, $_))[0] } @dir;
   foreach my $listFile ( sort _ipdrsort grep { $_ ne '.' && $_ ne '..' } @dir  ) {
       push(@$fileArray, $listFile) if $listFile;
   }

   foreach my $f ( @$fileArray ) {
       if ( ! grep { $_ eq $f } @old_files ) {
            if ( $fcounter < $maxFilesPerRun ) {
               if ( _ipdrstamp($f) < $lastAMAstamp && _ipdrstamp($f) > $lastAMAstamp - ( 1200 * $amaFilesPerRun ) ) {
                  # we have a new file...
                  #  $ftp->binary();
                  print "Downloading file: $ipdrpath/$f ... ";
                  move("$ipdrpath/$f", "$ipdrdir/$f");
                  print "Done.\n";
                  push ( @$newFileArr, $f );
                  $fcounter++;
               }
            }
       } else {
            print "Skipping already obtained file: $f\n";
       }

   }


   my $timestampStart = time();
   if ( $newFileArr ) {
      print "Processing IPDR files.... TIMESTAMP: $timestampStart\n";
   }


   foreach my $file ( @$newFileArr ) {
      my ( $data, $foo );
      open (IPDRFILE, "gzip -dc $ipdrdir/$file |") || die "Unable to open file '$file': $!\n";
      my $lineCount = 0;
      my $record;
      foreach my $line (<IPDRFILE>) {
         $lineCount++;
         if ( $line =~ /<IPDR>/ ) {
         # new record
         if ( $record ) {
                # we have a complete record. flush it to the database.
                # .. logic logic
                foreach my $key ( keys %$record ) {
                   $recordHash->{$record->{uniqueCallId}}->{$key} = $record->{$key};
                }
                $recordHash->{$record->{uniqueCallId}}->{ipdr} = 1;
                $record = {};
             }  
         } elsif ( $line !~ /IPDRDoc/ && $line !~ /xml/ ) {
             my ( $tag, $value  ) = $line =~ /<(.*)>(.*)</;
             # 2007-11-5T10:12:9Z
             $record->{$tag} = $value;
             if ( $tag =~ /EndTime/ ) {
                my ( $date, $time ) = split(/T/, $value);
                my ( $year, $month, $day ) = split(/-/, $date);
                my ( $hour, $minute, $second ) = split(/:/, $time);
                $record->{timestamp} = Date_SecsSince1970( $month, $day, $year, $hour, $minute, "0" );
             }
         }
      }
   }
}



################################################################################
################################################################################
################################################################################
##### DO NOT EDIT BELOW THIS LINE ##############################################
################################################################################
################################################################################
################################################################################



# defined( my $pid = fork);
my $pid = fork();


if ( $pid ) {
  # Watchdog routine
  # 
  while ( 1 ) {
     sleep 30;
     my $pid = `/bin/ps -eaf | grep perl | grep cs2k | awk '{print \$2}'`;
     my $parr = [ split(/\s+/, $pid) ];
     if ( scalar(@$parr) < 2 ) {
        # main isn't running...
        system("$launchpath");
        exit;
     }
  }
}



setsid;


$| = 1;


while ( 1 ) {
   $startEpoch = time();
   &_processIPDR();
   &_processAMA();
   &_flushData();
   my $timeNow = time();
   my $sleep = ( $startEpoch + $runInterval ) - $timeNow;
   if ( $sleep > 0 ) {
      print "Sleeping for $sleep\n";
      sleep $sleep;
   }
}
